#include "particle.h"
#include <GL/glut.h>
#include "particles.h"
#include "matrix.h"
#include "matrix3x3f.h"
#include <iostream>
#include <cfloat>
#include "lineSegment.h"
#include "lines.h"
using namespace std;

void Particle::draw(Particles * particles, Lines * lines, bool draw_shape_function) {
   // loop over support of particle
   const float h = getSupportRadius();
   float startx = floor(x - h);
   float endx   = ceil(x + h);
   float starty = floor(y - h);
   float endy   = ceil(y + h);

   // step size, only draw every step pixels
   float step = 1.0f;

   // we will store the values here
   // and scale them afterwards for nice rendering
   vector<float> values;
   float min_value = FLT_MAX;
   float max_value = -FLT_MAX;
   for (float x=startx; x<=endx; x+=step) {
      for (float y=starty; y<=endy; y+=step) {
         // create a temp particle
         Particle p(vector3f(x,y,0), h);
         // compute its neighbors
         p.setParticleNeighbors(particles, lines);
         // evaluate the shape function only for this index
         p.evaluateShapeFunctions(particles, getIndex());
         Neighbor * data = p.getNeighborWithIndex(getIndex());
         float value = (data==0?0:(draw_shape_function?data->shapeFunction:data->weight));
         if (value<min_value) min_value = value;
         if (value>max_value) max_value = value;
         values.push_back(value);
      }
   }
   // let's scale the shape function values so that they are
   // between 0 and 1 for nice rendering
   float scale = 1.0f/(max_value);
   for (int i=0; i<values.size(); i++) {
      values[i] = (values[i]) * scale;
   }

   glDisable(GL_POINT_SMOOTH);
   glDisable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glDisable(GL_DEPTH_TEST);
   glPointSize(step);
   glBegin(GL_POINTS);
   // actual rendering
   int index = 0;
   for (float x=startx; x<=endx; x+=step) {
      for (float y=starty; y<=endy; y+=step) {
         vector3f color((1-values[index])*240.0f, 1, 1);
         if (color.x>340) color.x= 340;
         color.HSVtoRGB();

         // also draw the iso-lines (every 0.1)
         int band = round(10.0f*values[index]);
         float fband = float(band)/10.0f;
         float diff = fabs(values[index]-fband);
         float alpha = diff>0.03?0:(0.03f-diff)/0.03f;
         alpha *= alpha; alpha *= alpha;
         if (fabs(values[index])>0.03f) glColor4f((1-alpha)*color.x,(1-alpha)*color.y,(1-alpha)*color.z,1);
         else glColor4f(color.x, color.y, color.z,1);
         glVertex2f(x,y);
         index++;
      }
   }
   glEnd();
}

void Particle::setParticleNeighbors(Particles * particles, Lines * lines) {
   // Simply loop over all particles and add the particles
   // for which I am in the support to my neighbor vectors.
   // Also compute an appropriate weight.
   const int nb = particles->nbParticles();
   for (int i=0; i<nb; i++) {
      Particle &pi = particles->getParticle(i);
      if (distance(pi) < pi.getSupportRadius()) {
         Neighbor n; n.index = i;
         n.weight = computeWeight(pi, lines);
         addNeighbor(n);
      }
   }
}

float Particle::computeWeight(const vector3f &p, Lines * lines) {
   // line segment that connects me with the given point
   LineSegment l(p,*this);

   float result_t;
   vector3f result_p;
   float dist_tip = 0;

   if (lines!=0) {
      for (int i=0; i<lines->nbLines(); i++) {
         LineSegment* l2 = lines->getLine(i);
         if (l2->intersection(l,result_t,result_p)) {
            if (l2->isTransparent()) dist_tip = l2->getTip().distance(result_p);
            else return 0;
         }
      }
   }
   const float h = getSupportRadius();
   const float inv_h = 1.0f/h;
   float w1 = (this->distance(p)*inv_h)+4*dist_tip*dist_tip*inv_h*inv_h; w1 = 1-w1*w1;
   float w = w1*w1*w1;
   if (w<0) w = 0;
   return w;


}

void Particle::evaluateShapeFunctions(Particles * particles, int onlyForIndex) {
   const int nb = nbNeighbors();

   // for convenience
   const float h = getSupportRadius();
   const float inv_h = 1.0f/h;

   // compute moment matrix
   MyMatrix M(3,4);
   for (int i=0; i<nb; i++) {
      Neighbor &n = getNeighbor(i);
      Particle &p = particles->getParticle(n.index);
      // weight based on distance to the neighbor
      const float w = n.weight;
      // translate and scale for better conditioning number
      const vector3f pp = (p - (*this))*inv_h;
      const float x = pp[0]; const float y = pp[1];
      // add to moment matrix
      M(0,0)+=w*y*y;
      M(1,1)+=w*x*x;
      M(2,2)+=w;
      M(0,1)+=w*x*y; M(1,0)+=w*y*x;
      M(0,2)+=w*y;   M(2,0)+=w*y;
      M(1,2)+=w*x;   M(2,1)+=w*x;
      // in the above formulation we use the linear basis
      // vector in 2D: [y x 1]
      
   }

   // shape functions should sum to one
   float test_sum_to_one = 0;
   // evaluate the shape functions
   for (int i=0; i<nb; i++) {
      Neighbor &n = getNeighbor(i);
      if (onlyForIndex!=-1 && n.index!=onlyForIndex) continue;
      Particle &p = particles->getParticle(n.index);
      // weight based on distance to the neighbor, computed above
      float w = n.weight;
      // translate and scale for better conditioning number
      const vector3f pp = (p - (*this))*inv_h;
      // evaluate final shape function for this neighbor
      MyMatrix M2(M);
      M2(0,3) = pp[1];
      M2(1,3) = pp[0];
      M2(2,3) = 1;
      M2.gausseidel();
      // thanks to the translation, we can just take the last element of the result
      float result = w*M2.result[2];
      n.shapeFunction = result;
      // shape functions should sum to one
      test_sum_to_one += result;
      // in the above formulation we use the linear basis
      // vector in 2D: [y x 1]
   }
   // this should output 1
   // cout << "   Shape functions sum to: " << test_sum_to_one << endl;
}

void Particle::evaluateGradientShapeFunctions(Particles * particles) {
   const int nb = nbNeighbors();

   // for convenience
   const float h = getSupportRadius();
   const float inv_h = 1.0f/h;

   // compute moment matrix and its spatial derivatives
   MyMatrix M(3,4);
   matrix3x3f Mx; Mx.zero();
   matrix3x3f My; My.zero();

   for (int i=0; i<nb; i++) {
      Neighbor &n = getNeighbor(i);
      Particle &p = particles->getParticle(n.index);
      // weight based on distance to the neighbor
      float w1 = (this->distance(p)*inv_h); w1 = 1-w1*w1;
      float w = w1*w1*w1;
      // gradient of the weight function
      vector3f gradient_w = (*this - p)*3*w1*w1*(-2*inv_h*inv_h);
      // translate and scale for better conditioning number
      const vector3f pp = (p - (*this))*inv_h;
      
      // for convenience
      const float x = pp[0]; const float y = pp[1];
      // add to moment matrix
      M(0,0)+=w*y*y;
      M(1,1)+=w*x*x;
      M(2,2)+=w;
      M(0,1)+=w*x*y; M(1,0)+=w*y*x;
      M(0,2)+=w*y;   M(2,0)+=w*y;
      M(1,2)+=w*x;   M(2,1)+=w*x;
      // in the above formulation we use the linear basis
      // vector in 2D: [y x 1]

      // for convenience
      const float wx = gradient_w[0]; const float wy = gradient_w[1];

      // spatial derivatives (negatives)
      Mx(0,0)-=wx*y*y;
      Mx(0,1)-=(wx*x*y - w*inv_h*y);
      Mx(0,2)-=wx*y;
      Mx(1,1)-=(wx*x*x-2*w*inv_h*x);
      Mx(1,2)-=(wx*x-w*inv_h);
      Mx(2,2)-=wx;

      My(0,0)-=wy*y*y-2*w*inv_h*y;
      My(0,1)-=wy*x*y-w*inv_h*x;
      My(0,2)-=wy*y-w*inv_h;
      My(1,1)-=wy*x*x;
      My(1,2)-=wy*x;
      My(2,2)-=wy;
      
   }

   // fill in symmetric elements
   for (int i=1; i<=2; i++) {
      for (int j=0; j<i; j++) {
         Mx(i,j) = Mx(j,i);
         My(i,j) = My(j,i);
      }
   }

   // compute inverted moment matrix
   matrix3x3f invM = M.invert();

   // compute spatial derivatives of inverted moment matrix
   matrix3x3f invMx = invM*Mx*invM;
   matrix3x3f invMy = invM*My*invM;

   // shape functions should sum to zero
   vector3f test_sum_to_zero(0,0,0);
   // evaluate the gradient of the shape functions
   for (int i=0; i<nb; i++) {
      Neighbor &n = getNeighbor(i);
      Particle &p = particles->getParticle(n.index);
      // weight based on distance to the neighbor
      float w1 = (this->distance(p)*inv_h); w1 = 1-w1*w1;
      float w = w1*w1*w1;
      // gradient of the weight function
      vector3f gradient_w = (*this - p)*3*w1*w1*(-2*inv_h*inv_h);
      // translate and scale for better conditioning number
      const vector3f pp = (p - (*this))*inv_h;
      
      // for convenience
      const float x = pp[0]; const float y = pp[1];
      const float wx = gradient_w[0]; const float wy = gradient_w[1];

      // dPhi/dx, first term (see slides)
      float resultx_1 = invM(2,2)*1;
      for (int k=0; k<2; k++) resultx_1+=invM(2,k)*pp[1-k];
      resultx_1*=wx;
      // dPhi/dx, second term (see slides)
      float resultx_2 = invMx(2,2)*1;
      for (int k=0; k<2; k++) resultx_2+=invMx(2,k)*pp[1-k];
      resultx_2*=w;
      // dPhi/dx, third term (see slides)
      float resultx_3 = invM(1,2)*(-inv_h);
      //for (int k=0; k<2; k++) resultx_3+=invM(1,k)*pp[1-k]*(-inv_h);
      resultx_3*=w;
      // final result for x
      float dPhi_dx = resultx_1+resultx_2+resultx_3;

      // dPhi/dy, first term (see slides)
      float resulty_1 = invM(2,2)*1;
      for (int k=0; k<2; k++) resulty_1+=invM(2,k)*pp[1-k];
      resulty_1*=wy;
      // dPhi/dy, second term (see slides)
      float resulty_2 = invMy(2,2)*1;
      for (int k=0; k<2; k++) resulty_2+=invMy(2,k)*pp[1-k];
      resulty_2*=w;
      // dPhi/dy, third term (see slides)
      float resulty_3 = invM(0,2)*(-inv_h);
      //for (int k=0; k<2; k++) resulty_3+=invM(0,k)*pp[1-k]*(-inv_h);
      resulty_3*=w;
      // final result for y
      float dPhi_dy = resulty_1+resulty_2+resulty_3;

      n.gradientShapeFunction.set(dPhi_dx, dPhi_dy, 0);
      //n.gradientShapeFunction.print();
      test_sum_to_zero+=n.gradientShapeFunction;
   }

   // this should output to 0 0 
   cout << "   Gradient shape functions sum to: " << test_sum_to_zero[0] << " " << test_sum_to_zero[1] << endl;
}


