#ifndef __PARTICLE_H__
#define __PARTICLE_H__

#include "vector3f.h"
#include "matrix3x3f.h"
#include "neighbor.h"
#include <vector>
using namespace std;

class Particles;
class Lines;

/* A particle is the basic primitive in the
 * meshless simulation framework. It is defined
 * by a position and a support radius
 *
 * Each particle keeps a list of its neighbors.
 */
class Particle : public vector3f {
   public:
      // Create a particle at the given position with the given support
      // radius. The index should be a unique identifier for this particle.
      Particle(const vector3f &pos, const float h, const int index=0) : 
         vector3f(pos), _h(h), _index(index) {}
     
      // Returns the unique ID for this particle.
      inline int getIndex() const { return _index; } 

      // Returns the support radius.
      inline float getSupportRadius() const { return _h; }

      // Number of neighbors of this particle.
      inline int nbNeighbors() const { return _neighbors.size(); }

      // Returns the index-th neighbor
      inline Neighbor& getNeighbor(const int index) { return _neighbors[index]; }

      // Returns the neighbor with the given id.
      // If no such neighbor is found, 0 is returned
      inline Neighbor* getNeighborWithIndex(const int index) { 
         for (int i=0; i<_neighbors.size(); i++) {
            if (_neighbors[i].index == index) return &_neighbors[i]; 
         }
         return 0;
      }

      // Add a new neighbor
      inline void addNeighbor(const Neighbor &n) { _neighbors.push_back(n); }

      // Computes all neighboring particles that have a support
      // that overlaps with this particle
      void setParticleNeighbors(Particles * particles, Lines * lines);   

      // Evaluate all shape functions for the neighbors of this particle
      // if onlyForIndex equals -1. Otherwise, only evaluate the shape function
      // of the neighbor with id equal to onlyForIndex.
      void evaluateShapeFunctions(Particles * particles, int onlyForIndex=-1);

      // Evaluate all the spatial derivatives of the shape functions for all neighbors.
      void evaluateGradientShapeFunctions(Particles * particles);

      // Draw the shape function or weight for this particle.
      // If draw_shape_function==true, the shape function is drawn,
      // otherwise the weight function is drawn.
      // This can be slow, since the shape function is evaluated
      // for every pixel in the support of this particle.
      void draw(Particles * particles, Lines * lines, bool draw_shape_function);

   private:
      // Helper function to compute the weight between this
      // particle and the given point p.
      float computeWeight(const vector3f &p, Lines * lines);

      int _index; // index (this should be a unique id for me)
      float _h; // particle support radius
      vector<Neighbor> _neighbors; // particle neighbors
};

#endif
